(* Ich, Doktor Josef Grosch, Informatiker, 28.10.1993 *)

MODULE lpp;

FROM rSystem	IMPORT GetArgCount, GetArgument, rSystem, rExit;
FROM rMemory	IMPORT MemoryUsed;
FROM IO		IMPORT StdOutput, StdError, WriteOpen, WriteClose, CloseIO,
			WriteI, WriteS, WriteNl,
			tFile, ReadOpen, ReadClose, EndOfFile, ReadNl;
FROM Strings	IMPORT tString, AssignEmpty, Assign, Append, Concatenate,
			ReadL, WriteL, ArrayToString, StringToArray;
FROM Idents	IMPORT tIdent, NoIdent, MakeIdent, GetString, WriteIdent;
FROM Sets	IMPORT tSet, MakeSet, ReleaseSet, Include, Exclude, IsEmpty,
			Extract, Complement, Union, IsElement;
FROM Errors	IMPORT StoreMessages, WriteMessages;
FROM Scanner	IMPORT BeginFile, Attribute, CurArg, ArgCount;
FROM Parser	IMPORT Parser;
FROM Tree	IMPORT NoTree, tTree, Trace, Options, TreeRoot, QueryTree, ErrorCount,
			HeapUsed, ModuleName, f, BeginTree2, ForallClasses, SourceFile,
			WI, TypeNames, itTree, iNoTree, iModule, iInteger, IdentifyClass,
			ViewName;
FROM Semantics	IMPORT Semantics, TypeCount;
FROM Order	IMPORT Order;
FROM Dialog	IMPORT Dialog;

IMPORT GramMod;
IMPORT GramC, GramYacc;

VAR
   CgLib	,
   DefFile	,
   TreeFile	,
   EvalFile	,
   String	,
   PathS	: tString;
   Argument	,
   PathA	: ARRAY [0..255] OF CHAR;
   j		: SHORTCARD;
   ch		: CHAR;
   AllOptions	: tSet;
   HasArgs	: BOOLEAN;

PROCEDURE SmartOpen (FileName: tString): tFile;
   VAR PathS: tString;
   BEGIN
      IF IsElement (ORD ('7'), Options) THEN
	 ArrayToString	("yy", PathS);
	 Concatenate	(PathS, FileName);
	 Append		(PathS, 0C);
	 StringToArray	(PathS, PathA);
      ELSE
	 Append		(FileName, 0C);
	 StringToArray	(FileName, PathA);
      END;
      RETURN WriteOpen	(PathA);
   END SmartOpen;

PROCEDURE SmartClose (FileName: tString; f: tFile);
   VAR PathS: tString;
   BEGIN
      WriteClose	(f);
      IF IsElement (ORD ('7'), Options) THEN
	 ArrayToString	("gmdupd ", PathS);
	 Concatenate	(PathS, FileName);
	 ArrayToString	(" yy", String);
	 Concatenate	(PathS, String);
	 Concatenate	(PathS, FileName);
	 Append		(PathS, 0C);
	 StringToArray	(PathS, PathA);
	 j := rSystem	(PathA);
      END;
   END SmartClose;

BEGIN
   AssignEmpty (CgLib);
   SourceFile [0] := 0C;
   ArgCount := GetArgCount () - 1;
   HasArgs := FALSE;

   CurArg := 0;
   LOOP
      INC (CurArg);
      IF CurArg > ArgCount THEN EXIT; END;
      GetArgument (CurArg, Argument);
      IF Argument [0] = '-' THEN 
	 IF Argument [1] = 'l' THEN 
	    AssignEmpty (CgLib);
	    j := 2;
	    LOOP
	       ch := Argument [j];
	       IF ch = 0C THEN EXIT; END;
	       Append (CgLib, ch);
	       INC (j);
	    END;
	    Append (CgLib, '/');
	 ELSE
	    j := 0;
	    LOOP
	       INC (j);
	       ch := Argument [j];
	       IF ch = 0C THEN
		  EXIT;
	       ELSIF ch = '?' THEN
		  Include (Options, ORD ('h'));
	       ELSE
		  Include (Options, ORD (ch));
	       END;
	    END;
	    HasArgs := TRUE;
	 END;
      ELSIF Argument [0] = '?' THEN 
	 Include (Options, ORD ('h'));
	 HasArgs := TRUE;
      ELSE
	 j := 0;
	 REPEAT
            ch := Argument [j];
	    SourceFile [j] := ch;
	    INC (j);
	 UNTIL ch = 0C;
	 EXIT;
      END;
   END;

   IF NOT HasArgs THEN				(* {} -> {x, z}		*)
      Include (Options, ORD ('x'));
      Include (Options, ORD ('z'));
   END;

   IF IsElement (ORD ('x'), Options) OR		(* {x, z, u} -> . = Order, B *)
      IsElement (ORD ('z'), Options) OR
      IsElement (ORD ('u'), Options) THEN
      Include (Options, ORD ('.'));
      Include (Options, ORD ('B'));
   END;

   IF IsElement (ORD ('h'), Options) THEN
      Exclude (Options, ORD ('h'));
      f := StdOutput;
      !!
      !usage: lpp [-options] [files]!
      !!
      ! x generate scanner specification!
      ! z generate parser  specification for lalr!
      ! u generate parser  specification for yacc!
      ! v omit actions in the generated parser specifications!
      ! j allow undefined node types; define implicitly as terminals!
      ! c generate C code (default: Modula-2)!
      ! h print help information for tree/graph module!
      ! 1 print inserted copy rules!
      ! 2 print inherited attribute computation rules!
      ! 6 generate # line directives!
      ! 7 touch output files only if necessary!
      ! 8 report storage consumption!
   END;

   IF NOT IsEmpty (Options) THEN

   IF IsElement (ORD ('c'), Options) THEN
      ArrayToString	("int", String);
   ELSE
      ArrayToString	("INTEGER", String);
   END;
      iInteger := MakeIdent (String);

      IF SourceFile [0] # 0C THEN
	 BeginFile (SourceFile);
	 ArrayToString (SourceFile, String);
	 Attribute.Position.File := MakeIdent (String);
      END;
      StoreMessages	(TRUE);
      ErrorCount := Parser ();
      BeginTree2;
      Semantics		(TreeRoot);
      IF ErrorCount > 0 THEN WriteMessages (StdError); CloseIO; rExit (1); END;
   END;

   IF IsElement (ORD ('x'), Options) THEN
      IF TreeRoot^.Ag.ScannerName = NoIdent THEN
	 ArrayToString	("Scanner", PathS);
      ELSE
	 GetString (TreeRoot^.Ag.ScannerName, PathS);
      END;
      ArrayToString	(".rpp", String);
      Concatenate	(PathS, String);
      f := SmartOpen	(PathS);
   IF IsElement (ORD ('c'), Options) THEN
      GramC.ScanSpec	(TreeRoot);
   ELSE
      GramMod.ScanSpec	(TreeRoot);
   END;
      SmartClose	(PathS, f);
   END;

   IF IsElement (ORD ('z'), Options) THEN
      GetString		(TreeRoot^.Ag.ParserName, ModuleName);
      Assign		(PathS, ModuleName);
      ArrayToString	(".lalr", String);
      Concatenate	(PathS, String);
      f := SmartOpen	(PathS);
   IF IsElement (ORD ('c'), Options) THEN
      GramC.ParsSpec	(TreeRoot);
   ELSE
      GramMod.ParsSpec	(TreeRoot);
   END;
      SmartClose	(PathS, f);
   ELSIF IsElement (ORD ('u'), Options) THEN
      GetString		(TreeRoot^.Ag.ParserName, ModuleName);
      Assign		(PathS, ModuleName);
      ArrayToString	(".y", String);
      Concatenate	(PathS, String);
      f := SmartOpen	(PathS);
      GramYacc.ParsSpec	(TreeRoot);
      SmartClose	(PathS, f);
   END;

   IF IsElement (ORD ('.'), Options) THEN
      Order		(TreeRoot);
   END;

      WriteMessages	(StdError);

   IF IsElement (ORD ('Q'), Options) THEN
      QueryTree		(TreeRoot);
   END;

   IF IsElement (ORD ('J'), Options) THEN
      Dialog		(TreeRoot);
   END;

      IF ErrorCount > 0 THEN CloseIO; rExit (1); END;

   IF IsElement (ORD ('8'), Options) THEN
      WriteNl (StdError);
      WriteS  (StdError, "Memory"); WriteI (StdError, MemoryUsed, 8);
      WriteS  (StdError, "  Heap"); WriteI (StdError, HeapUsed  , 8);
      WriteNl (StdError);
   END;
   CloseIO; rExit (0);
END lpp.
