#include "SYSTEM_.h"

#ifndef DEFINITION_IO
#include "IO.h"
#endif

#ifndef DEFINITION_Sets
#include "Sets.h"
#endif

#ifndef DEFINITION_Positions
#include "Positions.h"
#endif

#ifndef DEFINITION_Errors
#include "Errors.h"
#endif


typedef struct S_1 {
    CHAR A[256 + 1];
} tArray;
static CARDINAL ErrorCount;
static void WriteErrorMessage ARGS((CARDINAL ErrorCode, CARDINAL ErrorClass, Positions_tPosition Position));

CARDINAL Errors_NumberOfErrors
# ifdef __STDC__
()
# else
()
# endif
{
  return ErrorCount;
}

static void WriteErrorMessage
# ifdef __STDC__
(CARDINAL ErrorCode, CARDINAL ErrorClass, Positions_tPosition Position)
# else
(ErrorCode, ErrorClass, Position)
CARDINAL ErrorCode, ErrorClass;
Positions_tPosition Position;
# endif
{
  INC(ErrorCount);
  Positions_WritePosition((rSystem_tFile)IO_StdError, Position);
  IO_WriteS((rSystem_tFile)IO_StdError, (STRING)": ", 2L);
  switch (ErrorClass) {
  case Errors_Fatal:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Fatal      ", 11L);
    break;
  case Errors_Restriction:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Restriction", 11L);
    break;
  case Errors_Error:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Error      ", 11L);
    break;
  case Errors_Warning:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Warning    ", 11L);
    break;
  case Errors_Repair:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Repair     ", 11L);
    break;
  case Errors_Note:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Note       ", 11L);
    break;
  case Errors_Information:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Information", 11L);
    break;
  default :
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)"Error class: ", 13L);
    IO_WriteI((rSystem_tFile)IO_StdError, (LONGINT)ErrorClass, 0L);
    break;
  }
  switch (ErrorCode) {
  case Errors_NoText:;
    break;
  case Errors_SyntaxError:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" syntax error", 13L);
    break;
  case Errors_ExpectedTokens:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" expected tokens:", 17L);
    break;
  case Errors_RestartPoint:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" restart point", 14L);
    break;
  case Errors_TokenInserted:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" token inserted :", 17L);
    break;
  case Errors_ReadParseTable:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" error reading Pars.Tab", 23L);
    break;
  case Errors_IllegalChar:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" illegal character", 18L);
    break;
  case Errors_UnclosedComment:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" unclosed comment", 17L);
    break;
  case Errors_UnclosedString:;
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" unclosed string", 16L);
    break;
  default :
    IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" error code: ", 13L);
    IO_WriteI((rSystem_tFile)IO_StdError, (LONGINT)ErrorCode, 0L);
    break;
  }
}

void Errors_ErrorMessage
# ifdef __STDC__
(CARDINAL ErrorCode, CARDINAL ErrorClass, Positions_tPosition Position)
# else
(ErrorCode, ErrorClass, Position)
CARDINAL ErrorCode, ErrorClass;
Positions_tPosition Position;
# endif
{
  WriteErrorMessage(ErrorCode, ErrorClass, Position);
  IO_WriteNl((rSystem_tFile)IO_StdError);
  if (ErrorClass == Errors_Fatal) {
    IO_CloseIO();
    exit(1);
  }
}

void Errors_ErrorMessageI
# ifdef __STDC__
(CARDINAL ErrorCode, CARDINAL ErrorClass, Positions_tPosition Position, CARDINAL InfoClass, ADDRESS Info)
# else
(ErrorCode, ErrorClass, Position, InfoClass, Info)
CARDINAL ErrorCode, ErrorClass;
Positions_tPosition Position;
CARDINAL InfoClass;
ADDRESS Info;
# endif
{
  INTEGER *PtrToInteger;
  SHORTCARD *PtrToShort;
  LONGINT *PtrToLong;
  REAL *PtrToReal;
  BOOLEAN *PtrToBoolean;
  CHAR *PtrToCharacter;
  tArray *PtrToArray;

  WriteErrorMessage(ErrorCode, ErrorClass, Position);
  IO_WriteS((rSystem_tFile)IO_StdError, (STRING)" ", 1L);
  switch (InfoClass) {
  case Errors_Integer:;
    PtrToInteger = (LONGINT *)Info;
    IO_WriteI((rSystem_tFile)IO_StdError, *PtrToInteger, 0L);
    break;
  case Errors_Short:;
    PtrToShort = (SHORTCARD *)Info;
    IO_WriteI((rSystem_tFile)IO_StdError, (LONGINT)(*PtrToShort), 0L);
    break;
  case Errors_Long:;
    PtrToLong = (LONGINT *)Info;
    IO_WriteI((rSystem_tFile)IO_StdError, *PtrToLong, 0L);
    break;
  case Errors_Real:;
    PtrToReal = (REAL *)Info;
    IO_WriteR((rSystem_tFile)IO_StdError, *PtrToReal, 1L, 10L, 1L);
    break;
  case Errors_Boolean:;
    PtrToBoolean = (BOOLEAN *)Info;
    IO_WriteB((rSystem_tFile)IO_StdError, *PtrToBoolean);
    break;
  case Errors_Character:;
    PtrToCharacter = (CHAR *)Info;
    IO_WriteC((rSystem_tFile)IO_StdError, *PtrToCharacter);
    break;
  case Errors_Array:;
    PtrToArray = (tArray *)Info;
    IO_WriteS((rSystem_tFile)IO_StdError, (*PtrToArray).A, 257L);
    break;
  default :
    break;
  }
  IO_WriteNl((rSystem_tFile)IO_StdError);
  if (ErrorClass == Errors_Fatal) {
    IO_CloseIO();
    exit(1);
  }
}

void BEGIN_Errors()
{
  static BOOLEAN has_been_called = FALSE;
  if (has_been_called) return;
  has_been_called = TRUE;
    BEGIN_Positions();
    BEGIN_IO();
    BEGIN_Sets();
    BEGIN_Positions();

    ErrorCount = 0;
}
